/*-----------------------------------------------------------------------

File  : cte_replace.c

Author: Stephan Schulz

Contents
 
  Functions for replacing and rewriting of terms.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon Jan 26 19:16:16 MET 1998
    new

-----------------------------------------------------------------------*/

#include "cte_replace.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: TermAddRWLink()
//
//   Add a rewrite link from term to replace, induced by demod. Note:
//   If demod is REWRITE_AT_SUBTERM, actual rewriting happened at a
//   subterm.
//
// Global Variables: -
//
// Side Effects    : As described
//
/----------------------------------------------------------------------*/

void TermAddRWLink(Term_p term, Term_p replace, long demod)
{
   assert(term);
   assert(replace);
   assert(term!=replace);

   if(TermIsTopRewritten(term))
   {
      printf("Bug: Demod = %ld\n", demod);
   }
   assert(!TermIsTopRewritten(term));

   TermCellSetProp(term, TPIsRewritten);
   TermRWReplaceField(term) = replace;
   TermRWDemodField(term) = demod;
}


/*-----------------------------------------------------------------------
//
// Function: TermFollowRWChain()
//
//   Return the last term in an existing rewrite link chain.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Term_p TermFollowRWChain(Term_p term)
{
   assert(term);

   /* printf("Starting chain\n"); */
   while(TermIsRewritten(term))
   {
      term = TermRWReplaceField(term);
      /* printf("Following chain\n"); */
      assert(term);
   }
   return term;
}


/*-----------------------------------------------------------------------
//
// Function: TermFollowTopRWChain()
//
//   Return the last term in an existing rewrite link chain, following
//   only top rewrite links.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Term_p TermFollowTopRWChain(Term_p term)
{
   assert(term);

   /* printf("Starting chain\n"); */
   while(TermIsTopRewritten(term))
   {
      term = TermRWReplaceField(term);
      /* printf("Following chain\n"); */
      assert(term);
   }
   return term;
}


/*-----------------------------------------------------------------------
//
// Function: TBTermNormalize()
//
//   Given a term node in the term bank, return the term node
//   resulting by applying all existing rewrite links (top-down) in
//   the term and it's subterms. Do I really need this?
//
// Global Variables: -
//
// Side Effects    : Memory operations, may generate new term bank
//                   nodes, may introduce new rewrite links.
//
/----------------------------------------------------------------------*/

Term_p TBTermNormalize(TB_p bank, Term_p term)
{
   int i;   
   bool arg_changed = false;
   Term_p new_term;
   
   assert(bank);
   assert(term);

   if(TermIsVar(term))
   {
      assert(!TermIsRewritten(term));
      return term;
   }
	 
   term = TermFollowRWChain(term);
   new_term = TermTopCopy(term);
   for(i=0; i<term->arity; i++)
   {
      new_term->args[i] = TBTermNormalize(bank, term->args[i]);
      if(new_term->args[i]!=term->args[i])
      {
	 arg_changed = true;
      }
   }
   if(arg_changed)
   {
      new_term = TBTermtopInsert(bank, new_term);
      TermAddRWLink(term, new_term, REWRITE_AT_SUBTERM);
      term = new_term;
   }
   else
   {
      TermTopFree(new_term);
   }
   return term;
}


/*-----------------------------------------------------------------------
//
// Function: TermComputeRWSequence()
//
//   Given two terms from and two, connected by a rewrite chain, push
//   a sequence of clause idents onto the stack such that they
//   represent a rewrite chain transforming to into from. Returns true
//   if the chain has length 0.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool TermComputeRWSequence(PStack_p stack, Term_p from, Term_p to)
{
   bool   res = false;
   long   demod_id;
   Term_p tmp;

   while(from != to)
   {
      assert(TermIsRewritten(from));
      demod_id = TermRWDemod(from) ;
      tmp = TermRWReplaceField(from);
      if(demod_id == REWRITE_AT_SUBTERM)
      {
	 int i;
	 assert(from->f_code == tmp->f_code);
	 assert(from->arity&&from->args&&tmp->args);

	 for(i=0; i<from->arity; i++)
	 {
	    TermComputeRWSequence(stack, from->args[i], tmp->args[i]);
	 }
      }
      else
      {
	 PStackPushInt(stack, demod_id);
      }
      from = tmp;
      assert(from);
      res = true;
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: TBTermPosReplace()
//
//   Create a new term by replacing the subterm designated by pos with
//   repl and insert it into the term bank. Return pointer to the new
//   term. The superterm is implicit in the position (or, in the case
//   of the empty position, is unnecessary). Does not free any terms -
//   if necessary, this is the responsibility of the calling
//   functions.
//
//   Note that this function may dereference terms more than once,
//   which will lead to strange behaviour if instantiated variables
//   appear in the term...---> This has been fixed. Each term in repl
//   and its superterms should now be dereferenced only once.
//
// Global Variables: -
//
// Side Effects    : Memory management, changes term bank
//
/----------------------------------------------------------------------*/

Term_p TBTermPosReplace(TB_p bank, Term_p repl, TermPos_p pos,
			DerefType deref)
{
   Term_p        handle, old;
   int           subscript;
   PStackPointer i;
   PStack_p      store = PStackAlloc();

   DEBUGMARK(RPL_TERM_WATCH, "TBTermPosReplace()...\n");

   assert(bank);
   assert(repl);
   assert(pos);
   
   i = PStackGetSP(pos);
   
   while(i)
   {
      i--;
      subscript = PStackElementInt(pos, i);
      assert(i);
      i--;
      
      old = PStackElementP(pos, i);
      handle = TermTopCopy(old);
      assert(handle->arity > subscript);
      handle->args[subscript] = repl;
      PStackPushP(store, handle);
      repl = handle;
   }
   repl = TBInsertNoProps(bank, repl, deref);

   while(!PStackEmpty(store))
   {
      handle = PStackPopP(store);
      TermTopFree(handle);
   }
   PStackFree(store);

   DEBUGMARK(RPL_TERM_WATCH, "...TBTermPosReplace()\n");
   return repl;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


